/***************************************************************************
*   Copyright (C) 2001 by Rick L. Vinyard, Jr.                            *
*   rvinyard@cs.nmsu.edu                                                  *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU Lesser General Public License as        *
*   published by the Free Software Foundation version 2.1.                *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU Lesser General Public      *
*   License along with this library; if not, write to the                 *
*   Free Software Foundation, Inc.,                                       *
*   51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA              *
***************************************************************************/
#include "factory.h"

#include "bit-config.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <iostream>
#include <fstream>
#include <algorithm>
#include <cstring>

namespace bit
  {

  std::vector<std::string> xml_file_load_path;
  bool xml_file_load_path_initialized = false;

  void initialize_xml_file_load_path()
  {
    xml_file_load_path.clear();
    xml_file_load_path.push_back("");
    xml_file_load_path.push_back( "/usr/share/" BIT_PACKAGE_NAME_RELEASE "/");
    xml_file_load_path.push_back( "/usr/local/share/" BIT_PACKAGE_NAME_RELEASE "/");
  }

  RecordBase::pointer xml_load_file(const std::string& filename)
  {
    xmlpp::Document* document;
    xmlpp::Element* element;
    bool file_found = false;
    int last_position = 0;
    std::string file_with_path;
    struct stat statbuffer;

    xmlpp::DomParser parser;

    if (!xml_file_load_path_initialized)
      initialize_xml_file_load_path();

    for ( std::vector<std::string>::iterator i = xml_file_load_path.begin(); i != xml_file_load_path.end(); i++)
      {
        file_with_path = *i + filename;

        // Test for file existence
        if ( stat(file_with_path.c_str(), &statbuffer) == 0 )
          // Test whether the file is a regular file
          if (S_ISREG(statbuffer.st_mode) )
            // Make sure there is at least read access
            if ( access(file_with_path.c_str(), R_OK) == 0 )
              {
                // At this point we'll try reading the file
                //                 parser.set_validate();
                parser.set_substitute_entities();
                parser.parse_file(file_with_path);
                file_found = true;
                break;
              }
      }

    if (!file_found)
      {
        std::string s = "bit: Cannot load xml specification file ";
        s += filename;
        throw std::logic_error(s);
      }

    document = parser.get_document();
    if (document != NULL)
      {
        element = document->get_root_node();
        if (element != NULL)
          return xml_read_record(*element, last_position);
      }

    return RecordBase::pointer();
  }

  RecordBase::pointer xml_load_memory(const std::string& string)
  {
    xmlpp::Document* document;
    xmlpp::Element* element;
    xmlpp::DomParser parser;
    int last_position = 0;

    //                 parser.set_validate();
    parser.set_substitute_entities();
    parser.parse_memory(string);

    document = parser.get_document();
    if (document != NULL)
      {
        element = document->get_root_node();
        if (element != NULL)
          return xml_read_record(*element, last_position);
      }
    return RecordBase::pointer();
  }

  RecordBase::pointer xml_load_memory_raw(const unsigned char *contents, size_t bytes_count)
  {
    xmlpp::Document* document;
    xmlpp::Element* element;
    xmlpp::DomParser parser;
    int last_position = 0;

    //                 parser.set_validate();
    parser.set_substitute_entities();
    parser.parse_memory_raw(contents, bytes_count);

    document = parser.get_document();
    if (document != NULL)
    {
      element = document->get_root_node();
      if (element != NULL)
        return xml_read_record(*element, last_position);
    }
    return RecordBase::pointer();
  }

  RecordBase::pointer xml_load_stream(std::istream &in)
  {
    xmlpp::Document* document;
    xmlpp::Element* element;
    xmlpp::DomParser parser;
    int last_position = 0;

    //                 parser.set_validate();
    parser.set_substitute_entities();
    parser.parse_stream(in);

    document = parser.get_document();
    if (document != NULL)
    {
      element = document->get_root_node();
      if (element != NULL)
        return xml_read_record(*element, last_position);
    }
    return RecordBase::pointer();
  }

  RecordBase::pointer xml_read_record(xmlpp::Element& xml_element, int& last_position, size_t length_units, size_t offset_units)
  {
    RecordBase::pointer record;
    std::string name;
    int offset = last_position;
    size_t vector_size = 0;
    bool is_vector = false;

    xmlpp::Attribute* attribute;

    if ( xml_element.get_name() != "record" )
      {
        std::string s = "bit: Loading xml, expected 'record' element but found " + xml_element.get_name();
        throw std::logic_error(s);
      }

    attribute = xml_element.get_attribute("units");
    if (attribute != NULL)
      offset_units = length_units = get_units( attribute->get_value() );

    attribute = xml_element.get_attribute("start_units");
    if (attribute != NULL)
      offset_units = get_units( attribute->get_value() );

    attribute = xml_element.get_attribute("length_units");
    if (attribute != NULL)
      length_units = get_units( attribute->get_value() );

    attribute = xml_element.get_attribute("name");
    if (attribute != NULL)
      name = attribute->get_value();

    attribute = xml_element.get_attribute("start");
    if (attribute != NULL)
      // here we have an explicit offset, so we will simply set it to whatever it is
      offset = atoi( attribute->get_value().c_str() );
    else
      {
        // here, we have an implied offset, so we will keep the offset units the same if possible,
        // but if forced to we will resort to bits on the offset
        if (offset%offset_units == 0)
          offset /= offset_units;
        else
          offset_units = 1;
      }

    attribute = xml_element.get_attribute("vector_size");
    if (attribute != NULL)
      {
        vector_size = atoi( attribute->get_value().c_str() );
        is_vector = true;
      }

    if (is_vector)
      record = RecordVector::create(name, vector_size, length_units, offset, offset_units);
    else
      record = Record::create(name, length_units, offset, offset_units);

    xmlpp::Node::NodeList children;
    xmlpp::Node::NodeList::iterator iter;
    xmlpp::Element* child;
    int last_child_position = 0;

    children = xml_element.get_children();
    for (iter = children.begin(); iter != children.end(); iter++)
      {
        try
          {
            child = dynamic_cast<xmlpp::Element*>(*iter);
          }
        catch (std::bad_cast )
          {
            continue;
          }

        if (child == NULL)
          continue;

        if (child->get_name() == "description")
          {
            std::string description;
            if ( child->has_child_text() )
              {
                description = child->get_child_text()->get_content();
                while (isspace(description[description.size()-1]))
                  description.erase(description.size()-1);
                record->set_description(description);
              }
          }
        else if (child->get_name() == "record")
          {
            RecordBase::pointer ptr = xml_read_record(*child, last_child_position, length_units, offset_units);
            if (ptr) {
              record->add_field(ptr);
              // If the record has a default type, and the child doesn't have an override, set the child
              if (record->type().type() != TYPE_NONE && ptr->type().type() == TYPE_NONE)
                ptr->set_type( record->type() );
            }
          }
        else if (child->get_name() == "field")
          {
            Field::pointer ptr = xml_read_field(*child, last_child_position, length_units, offset_units);
            if (ptr) {
              record->add_field(ptr);

              // If the record has a default type, and the child doesn't have an override, set the child
              if (record->type().type() != TYPE_NONE && ptr->type().type() == TYPE_NONE) {
                ptr->set_type( record->type() );
              }
            }
          }
        else if (child->get_name() == "skip")
          {
            xml_read_skip(*child, last_child_position, length_units, offset_units);
          }
        else if (child->get_name() == "type")
            record->set_type( xml_read_type(*child) );
      }

    last_position = record->offset(BITS) + record->length(BITS);
    return record;
  }

  Field::pointer xml_read_field(xmlpp::Element& xml_element, int& last_position, size_t length_units, size_t offset_units)
  {
    Field::pointer field;
    std::string name;
    size_t length = 0;
    int offset = last_position;
    size_t vector_size = 0;
    bool is_vector = false;

    xmlpp::Attribute* attribute;

    if ( xml_element.get_name() != "field" )
      {
        std::string s = "bit: Loading xml, expected 'field' element but found " + xml_element.get_name();
        throw std::logic_error(s);
      }

    attribute = xml_element.get_attribute("units");
    if (attribute != NULL)
      offset_units = length_units = get_units( attribute->get_value() );

    attribute = xml_element.get_attribute("start_units");
    if (attribute != NULL)
      offset_units = get_units( attribute->get_value() );

    attribute = xml_element.get_attribute("length_units");
    if (attribute != NULL)
      length_units = get_units( attribute->get_value() );

    attribute = xml_element.get_attribute("name");
    if (attribute != NULL)
      name = attribute->get_value();

    attribute = xml_element.get_attribute("length");
    if (attribute != NULL)
      length = atoi( attribute->get_value().c_str() );

    attribute = xml_element.get_attribute("start");
    if (attribute != NULL)
      // here we have an explicit offset, so we will simply set it to whatever it is
      offset = atoi( attribute->get_value().c_str() );
    else
      {
        // here, we have an implied offset, so we will keep the offset units the same if possible,
        // but if forced to we will resort to bits on the offset
        if (offset%offset_units == 0)
          offset /= offset_units;
        else
          offset_units = 1;
      }

    attribute = xml_element.get_attribute("vector_size");
    if (attribute != NULL)
      {
        vector_size = atoi( attribute->get_value().c_str() );
        is_vector = true;
      }


    if (is_vector)
      field = FieldVector::create(name, vector_size, length, length_units, offset, offset_units);
    else
      field = Field::create(name, length, length_units, offset, offset_units);

    xmlpp::Node::NodeList children;
    xmlpp::Node::NodeList::iterator iter;
    xmlpp::Element* child;


    children = xml_element.get_children();
    for (iter = children.begin(); iter != children.end(); iter++)
      {
        try
          {
            child = dynamic_cast<xmlpp::Element*>(*iter);
          }
        catch (std::bad_cast )
          {
            continue;
          }

        if (child == NULL)
          continue;

        if (child->get_name() == "description")
          {
            std::string description;
            if ( child->has_child_text() )
              {
                description = child->get_child_text()->get_content();
                while (isspace(description[description.size()-1]))
                  description.erase(description.size()-1);
                field->set_description(description);
              }
          }

          if (child->get_name() == "type")
          {
            field->set_type( xml_read_type(*child) );
          }
      }

    last_position = field->offset(BITS) + field->length(BITS);
    return field;
  }

  FieldType xml_read_type(xmlpp::Element& xml_element) {
    FieldType type;
    xmlpp::Node::NodeList children;
    xmlpp::Node::NodeList::iterator iter;
    xmlpp::Element* child;


    children = xml_element.get_children();
    for (iter = children.begin(); iter != children.end(); iter++)
    {
      try
      {
        child = dynamic_cast<xmlpp::Element*>(*iter);
      }
      catch (std::bad_cast )
      {
        continue;
      }

      if (child == NULL)
        continue;

      if (child->get_name() == "integer")
      {
        type.set_type(TYPE_INTEGER);

        xmlpp::Attribute* attribute;

        attribute = xml_element.get_attribute("sign");
        if (attribute != NULL) {
          if (attribute->get_value() == "unsigned")
            type.set_sign(SIGN_UNSIGNED);
          else if (attribute->get_value() == "ones_complement")
            type.set_sign(SIGN_ONES_COMPLEMENT);
          else if (attribute->get_value() == "twos_complement")
            type.set_sign(SIGN_TWOS_COMPLEMENT);
          else if (attribute->get_value() == "sign_bit")
            type.set_sign(SIGN_BIT);
        }

        attribute = xml_element.get_attribute("byte_order");
        if (attribute != NULL) {
          if (attribute->get_value() == "network")
            type.set_byte_order(BYTEORDER_NETWORK);
          else if (attribute->get_value() == "big_endian")
            type.set_byte_order(BYTEORDER_BIG_ENDIAN);
          else if (attribute->get_value() == "little_endian")
            type.set_byte_order(BYTEORDER_LITTLE_ENDIAN);
        }
      }

      if (child->get_name() == "floating")
      {
        type.set_type(TYPE_FLOATING);

        xmlpp::Attribute* attribute;

        attribute = xml_element.get_attribute("byte_order");
        if (attribute != NULL) {
          if (attribute->get_value() == "network")
            type.set_byte_order(BYTEORDER_NETWORK);
          else if (attribute->get_value() == "big_endian")
            type.set_byte_order(BYTEORDER_BIG_ENDIAN);
          else if (attribute->get_value() == "little_endian")
            type.set_byte_order(BYTEORDER_LITTLE_ENDIAN);
        }
      }

      if (child->get_name() == "ascii")
      {
        type.set_type(TYPE_ASCII);
      }

      if (child->get_name() == "bcd")
      {
        type.set_type(TYPE_BCD);
      }

      if (child->get_name() == "utf8")
      {
        type.set_type(TYPE_UTF8);
      }

    }
    return type;
  }

  void xml_read_skip(xmlpp::Element& xml_element, int& last_position, size_t length_units, size_t offset_units)
  {
    int start=0, length=0;

    xmlpp::Attribute* attribute;

    attribute = xml_element.get_attribute("start");
    if ( attribute != NULL )
      start = atoi( attribute->get_value().c_str() );

    attribute = xml_element.get_attribute("length");
    if ( attribute != NULL )
      length = atoi( attribute->get_value().c_str() );

    attribute = xml_element.get_attribute("units");
    if (attribute != NULL)
      offset_units = length_units = get_units( attribute->get_value() );

    attribute = xml_element.get_attribute("start_units");
    if (attribute != NULL)
      offset_units = get_units( attribute->get_value() );

    attribute = xml_element.get_attribute("length_units");
    if (attribute != NULL)
      length_units = get_units( attribute->get_value() );

    last_position = start * offset_units + length * length_units;
  }

  size_t get_units(std::string s)
  {
    size_t units=0;
    std::transform(s.begin(), s.end(), s.begin(), tolower);
    if ( s.find("bit") != std::string::npos )
      units = 1;
    else if ( s.find("octet") != std::string::npos )
      units = 8;
    else
      units = atoi( s.c_str() );
    return units;
  }

  std::string get_xml_file_as_string( const std::string & filename )
  {
    std::ifstream infile;
    size_t length;
    std::string retstring;
    std::string file_with_path;
    char* buffer;
    if (!xml_file_load_path_initialized)
      initialize_xml_file_load_path();

    for ( std::vector<std::string>::iterator i = xml_file_load_path.begin(); i != xml_file_load_path.end(); i++)
    {
      file_with_path = *i + filename;

      infile.open(file_with_path.c_str());

      if ( infile.is_open() ) {
        // At this point we'll try reading the file
        infile.seekg(0, std::ios::end);
        length = infile.tellg();
        infile.seekg(0, std::ios::beg);
        buffer = new char[length];
        infile.read(buffer,length);
        retstring = buffer;
        delete[] buffer;
        infile.close();
        return retstring;
      }
    }

    std::string s = "bit: Cannot load xml specification file ";
    s += filename;
    throw std::logic_error(s);
  }

}
